﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Controls;
using System.ComponentModel;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Animation;

namespace Vis2NodeTrix
{
    /// <summary>
    /// Gives an visual abstraction of the BaseNode
    /// </summary>
    public abstract class BaseNodePresenter : Control, INotifyPropertyChanged
    {
        private struct NameHandlerMap
        {
            public String name;
            public AnimationEventHandler handler;
        }

        private static List<String> nodeNames = new List<string>();
        private static List<NameHandlerMap> nameHandlerMap = new List<NameHandlerMap>();

        /// <summary>
        /// The graphpresenter in which the Node is shown
        /// </summary>
        protected GraphPresenter gp;
        private bool translateAnimationFirstFinished;
        private Point animationEndPoint;
        private AnimationType curAnimationType;
        private GroupNode animationToGroupGroupNode;

        /// <summary>
        /// Initializes a new instance of the BaseNodePresenter class
        /// </summary>
        /// <param name="_gp">The graphpresenter in which the Node is shown</param>
        public BaseNodePresenter(GraphPresenter _gp)
        {
            gp = _gp;
        }

        /// <summary>
        /// Initializes the eventHandlers for this class
        /// </summary>
        protected void initialiseHandler()
        {
            if (nodeNames.Contains(Data.Name))
            {
                NameHandlerMap nhm = nameHandlerMap.First(x => x.name == Data.Name);
                nameHandlerMap.Remove(nhm);
                Data.AnimationEvent -= nhm.handler;

                AnimationEventHandler handler = new AnimationEventHandler(Node_AnimationEvent);
                Data.AnimationEvent += handler;

                nhm.handler = handler;

                nameHandlerMap.Add(nhm);
            }
            else
            {
                AnimationEventHandler handler = new AnimationEventHandler(Node_AnimationEvent);
                Data.AnimationEvent += handler;

                NameHandlerMap nhm = new NameHandlerMap();
                nhm.name = Data.Name;
                nhm.handler = handler;

                nodeNames.Add(Data.Name);
                nameHandlerMap.Add(nhm);
            }
        }

        /// <summary>
        /// Returns the unerlaying node of this presenter
        /// </summary>
        public abstract BaseNode Data { get; }

        /// <summary>
        /// Occurs when the value of a Property is set
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Throws the PropertyChangedEvent
        /// </summary>
        /// <param name="_property">Name of the property</param>
        protected virtual void OnPropertyChanged(string _property)
        {
            if (this.PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(_property));
        }

        private void Node_AnimationEvent(object sender, AnimationEventArgs e)
        {
            if (e.Type == AnimationType.TRANSLATE_FROM_MATRIX && e.Values.Length == 2)
            {
                this.RenderTransform = new TranslateTransform();

                curAnimationType = e.Type;
                translateAnimationFirstFinished = false;
                animationEndPoint = new Point(Data.Pos.X + e.Values[0], Data.Pos.Y + e.Values[1]);
                this.Data.IsAnimating = true;

                ElasticEase ease = new ElasticEase();
                ease.EasingMode = EasingMode.EaseOut;
                ease.Oscillations = 3;
                ease.Springiness = 7;
                /*BounceEase ease = new BounceEase();
                ease.EasingMode = EasingMode.EaseIn;
                ease.Bounces = 3;*/
                /*BackEase ease = new BackEase();
                ease.EasingMode = EasingMode.EaseIn;
                ease.Amplitude = 0.8;*/
                /*CircleEase ease = new CircleEase();
                ease.EasingMode = EasingMode.EaseOut;*/


                DoubleAnimation aX = new DoubleAnimation();
                aX.From = 0;
                aX.To = e.Values[0];
                aX.Duration = new Duration(TimeSpan.FromSeconds(1.5));
                aX.EasingFunction = ease;
                aX.Completed += new EventHandler(TranslateAnimation_Completed);
                DoubleAnimation aY = new DoubleAnimation();
                aY.From = 0;
                aY.To = e.Values[1];
                aY.Duration = new Duration(TimeSpan.FromSeconds(1.5));
                aY.Completed += new EventHandler(TranslateAnimation_Completed);
                aY.EasingFunction = ease;

                this.RenderTransform.BeginAnimation(TranslateTransform.XProperty, aX);
                this.RenderTransform.BeginAnimation(TranslateTransform.YProperty, aY);
            }
            if (e.Type == AnimationType.TRANSLATE_TO_MATRIX && e.Values.Length == 2 && e.Data is GroupNode)
            {
                this.RenderTransform = new TranslateTransform();
                curAnimationType = e.Type;
                animationToGroupGroupNode = e.Data as GroupNode;
                this.Data.IsAnimating = true;

                PowerEase ease = new PowerEase();
                ease.EasingMode = EasingMode.EaseIn;
                ease.Power = 5.5;

                DoubleAnimation aX = new DoubleAnimation();
                aX.From = 0;
                aX.To = e.Values[0] - Data.Pos.X;
                aX.Duration = new Duration(TimeSpan.FromSeconds(1.5));
                aX.EasingFunction = ease;
                aX.Completed += new EventHandler(TranslateAnimation_Completed);
                DoubleAnimation aY = new DoubleAnimation();
                aY.From = 0;
                aY.To = e.Values[1] - Data.Pos.Y;
                aY.Duration = new Duration(TimeSpan.FromSeconds(1.5));
                aY.Completed += new EventHandler(TranslateAnimation_Completed);
                aY.EasingFunction = ease;

                this.RenderTransform.BeginAnimation(TranslateTransform.XProperty, aX);
                this.RenderTransform.BeginAnimation(TranslateTransform.YProperty, aY);
            }
        }

        private void TranslateAnimation_Completed(object sender, EventArgs e)
        {
            if (curAnimationType == AnimationType.TRANSLATE_FROM_MATRIX)
            {
                if (translateAnimationFirstFinished == true)
                {
                    this.Data.Pos = animationEndPoint;
                    this.RenderTransform = Transform.Identity;
                    translateAnimationFirstFinished = false;
                    curAnimationType = AnimationType.NONE;
                    this.Data.IsAnimating = false;
                }
                else
                    translateAnimationFirstFinished = true;
            }
            if (curAnimationType == AnimationType.TRANSLATE_TO_MATRIX)
            {
                if (translateAnimationFirstFinished == true)
                {
                    if (!gp.Nodes.Contains(animationToGroupGroupNode))
                    {
                        gp.Nodes.Add(animationToGroupGroupNode);
                    }

                    if (Data is SimpleNode)
                    {
                        animationToGroupGroupNode.AddNode(Data as SimpleNode);
                    }
                    else
                    {
                        GroupNode gn = Data as GroupNode;
                        while (gn.Nodes.Count > 0)
                        {
                            SimpleNode n = gn.Nodes.ElementAt(0);
                            gn.RemoveNode(n);
                            animationToGroupGroupNode.AddNode(n);
                        }
                    }

                    gp.Nodes.Remove(Data);

                    translateAnimationFirstFinished = false;
                    curAnimationType = AnimationType.NONE;
                    this.Data.IsAnimating = false;
                }
                else
                    translateAnimationFirstFinished = true;
            }
        }
    }
}
