﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using GraphVizNet;
using System.Collections.ObjectModel;

namespace Vis2NodeTrix
{
    /// <summary>
    /// Interaction logic for GraphPresenter.xaml
    /// </summary>
    public partial class GraphPresenter : UserControl
    {
        private VisualCollection Visuals;
        private ObservableCollection<BaseNodePresenter> NodePresenter;
        private ObservableCollection<EdgePresenter> EdgePresenter;
        private bool showLabels;
        private bool showTitles;

        private Point translation;
        private double zoom;

        //Drag
        //private Point startPoint;
        //private BaseNode draggedNode;

        /// <summary>
        /// Initializes a new instance of the GraphPresenter class
        /// </summary>
        public GraphPresenter()
        {
            InitializeComponent();
            Visuals = new VisualCollection(this);
            NodePresenter = new ObservableCollection<BaseNodePresenter>();
            EdgePresenter = new ObservableCollection<EdgePresenter>();
            showLabels = true;
            showTitles = true;
            Zoom = 1.0;
            Translation = new Point(0.0, 0.0);
        }

        /// <summary>
        /// A list of all Nodes in the graph
        /// </summary>
        public ObservableCollection<BaseNode> Nodes
        {
            get { return (ObservableCollection<BaseNode>)GetValue(NodesProperty); }
            set
            {
                SetValue(NodesProperty, value);
            }
        }

        /// <summary>
        /// Dependecy property for the Nodes collection
        /// </summary>
        public static readonly DependencyProperty NodesProperty =
            DependencyProperty.Register("Nodes", typeof(ObservableCollection<BaseNode>), typeof(GraphPresenter), new FrameworkPropertyMetadata() { BindsTwoWayByDefault = false, AffectsRender = true, AffectsArrange = true, AffectsMeasure = true, PropertyChangedCallback=new PropertyChangedCallback(OnNodesChanged) });

        /// <summary>
        /// Indicates if labels are shown or not
        /// </summary>
        public bool ShowLabels { get { return showLabels; } set { showLabels = value; UpdateVisuals(); } }

        /// <summary>
        /// Indicates if group titles are shown or not
        /// </summary>
        public bool ShowTitles { get { return showTitles; } set { showTitles = value; UpdateVisuals(); } }

        /// <summary>
        /// Translation of the graphpresenter
        /// </summary>
        public Point Translation { get { return translation; } set { translation = value; UpdateTransformation(); } }

        /// <summary>
        /// Zoom of the graphpresenter
        /// </summary>
        public double Zoom { get { return zoom; } set { zoom = value; UpdateTransformation(); } }

        private void UpdateTransformation()
        {
            TransformGroup tg = new TransformGroup();
            tg.Children.Add(new ScaleTransform(Zoom, Zoom));
            tg.Children.Add(new TranslateTransform(Translation.X, Translation.Y));
            this.RenderTransform = tg;
        }


        private static void OnNodesChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            if (o is GraphPresenter)
            {
                GraphPresenter gp = o as GraphPresenter;

                if (e.OldValue != null)
                    (e.OldValue as ObservableCollection<BaseNode>).CollectionChanged -= gp.Nodes_CollectionChanged;
                if (e.NewValue != null)
                    (e.NewValue as ObservableCollection<BaseNode>).CollectionChanged += gp.Nodes_CollectionChanged;

                gp.OnNodesChanged();
            }
        }

        void Nodes_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            OnNodesChanged();
        }

        /// <summary>
        /// To be called if the nodes changes
        /// </summary>
        public void OnNodesChanged()
        {
            CreateVisuals();
        }

        /// <summary>
        /// Recalculates the visuals and updates the rendering
        /// </summary>
        public void UpdateVisuals()
        {
            this.CreateVisuals();
            this.InvalidateMeasure();
        }

        private void CreateVisuals()
        {
            Visuals.Clear();
            NodePresenter.Clear();

            foreach (EdgePresenter ep in EdgePresenter)
                ep.RemoveMyHandler();

            EdgePresenter.Clear();

            foreach (BaseNode vn in Nodes)
            {
                BaseNodePresenter np = null;

                if (vn is SimpleNode)
                    np = new SimpleNodePresenter(vn as SimpleNode,this);
                else if (vn is GroupNode)
                    np = new GroupNodePresenter(vn as GroupNode,this);

                Visuals.Add(np);
                NodePresenter.Add(np);
            }

            foreach (BaseNode vn in Nodes)
            {
                foreach (VizEdge ve in vn.Connections)
                {
                    if (ve.Head == vn || (ve.Head as SimpleNode).IsGrouped && (ve.Head as SimpleNode).ParentNode == vn)
                    {
                        EdgePresenter ep = new EdgePresenter(ve);
                        ep.UpdateCurrentLength();
                        ep.UpdateLine(vn);
                        Visuals.Add(ep);
                        EdgePresenter.Add(ep);
                    }
                }
            }

            this.InvalidateMeasure();
        }

        /// <summary>
        /// Returns the border size for this control
        /// </summary>
        public static double Border { get { return 100.0; } }

        private double addX = 0.0, addY = 0.0;

        /// <summary>
        /// Returns the minimum coordinates of the nodes
        /// </summary>
        public Point minCoords { get; set; }

        /// <summary>
        /// Overrides the default measuring behaviour of the graphpresenter
        /// </summary>
        /// <param name="availableSize">The available size for this control</param>
        /// <returns>The size the control wants to have</returns>
        protected override Size MeasureOverride(Size availableSize)
        {
            double maxX = 0, maxY = 0;
            double minX = 0.0, minY = 0.0;

            foreach (BaseNodePresenter n in NodePresenter)
            {
                n.Measure(new Size(Double.PositiveInfinity, Double.PositiveInfinity));
                n.Data.Size = n.DesiredSize;

                double x = n.DesiredSize.Width + n.Data.Pos.X,
                    y = n.DesiredSize.Height + n.Data.Pos.Y;

                maxX = Math.Max(maxX, x);
                maxY = Math.Max(maxY, y);
                minX = Math.Min(minX, n.Data.Pos.X - n.DesiredSize.Width / 2.0);
                minY = Math.Min(minY, n.Data.Pos.Y - n.DesiredSize.Height / 2.0);
            }

            foreach (EdgePresenter e in EdgePresenter)
            {
                e.Measure(new Size(Double.PositiveInfinity, Double.PositiveInfinity));

                maxX = Math.Max(maxX, e.DesiredSize.Width);
                maxY = Math.Max(maxY, e.DesiredSize.Height);
            }

            if (minX < Border / 2.0)
                addX = Math.Abs(minX);
            if (minY < Border / 2.0)
                addY = Math.Abs(minY);

            return new Size(maxX + Border + addX, maxY + Border + addX);
        }

        /// <summary>
        /// Overrides the default layouting behaviour of the graphpresenter
        /// </summary>
        /// <param name="finalSize">The Size of the control should have</param>
        /// <returns>The final size of the control</returns>
        protected override Size ArrangeOverride(Size finalSize)
        {
            Size availableSize = this.DesiredSize;

            double maxY = 0.0, maxX = 0.0;
            double minX = double.PositiveInfinity, minY = double.PositiveInfinity;

            foreach (var node in NodePresenter)
            {
                var rect = new Rect();
                rect.Size = node.DesiredSize;
                rect.X = node.Data.Pos.X - (rect.Size.Width / 2) +(Border / 2.0) + addX;
                rect.Y = node.Data.Pos.Y - (rect.Size.Height / 2) +(Border / 2.0) + addY;

                maxY = Math.Max(maxY, rect.Y + rect.Size.Height);
                maxX = Math.Max(maxX, rect.X + rect.Size.Width);
                minX = Math.Min(minX, rect.X);
                minY = Math.Min(minY, rect.Y);

                if (addX != 0.0 || addY != 0.0)
                    node.Data.PosNoEvent = new Point(node.Data.Pos.X + addX, node.Data.Pos.Y + addY);

                node.Arrange(rect);
            }

            foreach (var a in EdgePresenter)
            {
                if (addX != 0.0 || addY != 0.0)
                {
                    for (int i = 0; i < a.Edge.Points.Count; i++)
                    {
                        a.Edge.Points[i] = new Point(a.Edge.Points[i].X + addX, a.Edge.Points[i].Y + addY);
                    }
                    a.MakeGeometry();
                }

                a.Arrange(new Rect(new Point(Border / 2.0, Border / 2.0), new Point(this.DesiredSize.Width + Border / 2.0, this.DesiredSize.Height + Border / 2.0)));
                a.DoBoundingBoxCorrection();

                maxY = Math.Max(maxY, this.DesiredSize.Height + Border / 2.0);
                maxX = Math.Max(maxX, this.DesiredSize.Width + Border / 2.0);
            }

            //this.UpdateTransformation();
            addX = addY = 0.0;
            minCoords = new Point(minX, minY);

            return new Size(Math.Max(finalSize.Width, maxX), Math.Max(finalSize.Height, maxY));
        }

        /// <summary>
        /// returns the amount of visuals
        /// </summary>
        protected override int VisualChildrenCount
        {
            get
            {
                return Visuals.Count;
            }
        }

        /// <summary>
        /// Returns one visual selected from its index
        /// </summary>
        /// <param name="index">Index of the visual</param>
        /// <returns>The selected visual</returns>
        protected override Visual GetVisualChild(int index)
        {
            return Visuals[index];
        }

        /// <summary>
        /// Calls the VizGraph library to layout the graph
        /// </summary>
        /// <param name="_nodes">Nodes to be layouted</param>
        public void LayoutGraph(IEnumerable<BaseNode> _nodes)
        {
            VizGraph graph = new VizGraph();
            graph.Type = VizGraphType.Graph;

            int i = 0;

            foreach (BaseNode n in _nodes)
            {
                if (n is GroupNode)
                {
                    n.Name = i.ToString().PadRight(6) + n.Name;
                    ++i;
                }
                n.Pos = new Point(0.0, 0.0);
                graph.Nodes.Add(n);
            }

            Dictionary<VizEdge, VizEdge> layoutEdgeDictionary = new Dictionary<VizEdge,VizEdge>();

            foreach (BaseNode n in _nodes)
            {
                foreach (var edge in n.Connections)
                {
                    SimpleNode node1 = edge.Tail as SimpleNode;
                    SimpleNode node2 = edge.Head as SimpleNode;

                    BaseNode bNode1 = node1;
                    if (node1.IsGrouped)
                        bNode1 = node1.ParentNode;

                    BaseNode bNode2 = node2;
                    if (node2.IsGrouped)
                        bNode2 = node2.ParentNode;

                    VizEdge searchEdge = graph.Edges.FirstOrDefault(p => (p.Head == bNode1 && p.Tail == bNode2) || (p.Head == bNode2 && p.Tail == bNode1));

                    if (searchEdge == null)
                    {
                        VizEdge ve = new VizEdge();
                        ve.Head = bNode1;
                        ve.Tail = bNode2;
                        graph.Edges.Add(ve);
                        layoutEdgeDictionary.Add(edge, ve);
                    }
                }
            }


            GraphEngine ge = new GraphEngine();
            ge.DotLayout(graph);
            graph.ConvertToLeftUpperCorner();

            foreach (BaseNode n in _nodes)
            {
                if (n is GroupNode)
                    n.Name = n.Name.Substring(6);
            }

            //Zurückkopieren
            foreach (var edgeRelation in layoutEdgeDictionary)
            {
                edgeRelation.Key.Points.Clear();
                foreach (Point p in edgeRelation.Value.Points)
                    edgeRelation.Key.Points.Add(p);
            }
        }

        /// <summary>
        /// Changes the order of the objects so that the forwarded object is on the bottom
        /// </summary>
        /// <param name="_presenter">Object which shall be braught to the bottom</param>
        public void BringToBottom(BaseNodePresenter _presenter)
        {
            if (Visuals.Count > 0 && Visuals[0] != _presenter)
            {
                Visuals.Remove(_presenter);
                Visuals.Insert(0, _presenter);
                this.InvalidateVisual();
            }
        }

        private void Title_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.ClickCount == 2)
            {
                GroupNode gn = (sender as TextBlock).Tag as GroupNode;
                string newTitle = InputBox.ShowDialog("Edit Title", "Title:", gn.Title);
                if (newTitle != String.Empty)
                    gn.Title = newTitle;
            }
        }
    }
}
