﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GraphVizNet;
using System.Collections.ObjectModel;
using System.Windows.Controls;
using System.ComponentModel;
using System.Windows;
using System.Windows.Media;

namespace Vis2NodeTrix
{
    /// <summary>
    /// Represents a groupnode (adjacencymatrix)
    /// </summary>
    public class GroupNode : BaseNode
    {
        private ObservableCollection<SimpleNode> nodes;
        private List<VizEdge> connections;
        private List<VizEdge> innerConnections;
        private string title;

        /// <summary>
        /// Gets or sets the title of the groupnode
        /// </summary>
        public string Title { get { return title; } set { title = value; OnPropertyChanged("Title"); } }

        /// <summary>
        /// Returns the connections to the groupnode
        /// </summary>
        public override List<VizEdge> Connections { get { return connections; } }
        /// <summary>
        /// Returns a list of all nodes in the groupnode
        /// </summary>
        public ObservableCollection<SimpleNode> Nodes { get { return nodes; } }
        /// <summary>
        /// Returns a list of all connections between nodes in the groupnode
        /// </summary>
        public List<VizEdge> InnerConnections { get { return innerConnections; } }

        /// <summary>
        /// Returns a list containing the the position of each node on the border
        /// </summary>
        public ObservableCollection<Point> NodesPosition { get; set; }

        /// <summary>
        /// Initializes a new instance of the GroupNode class
        /// </summary>
        public GroupNode()
        {
            nodes = new ObservableCollection<SimpleNode>();
            connections = new List<VizEdge>();
            innerConnections = new List<VizEdge>();
            NodesPosition = new ObservableCollection<Point>();

            Size = new System.Windows.Size(100, 100);
            Title = "";
            Name = "GroupNode";
        }

        /// <summary>
        /// Adds a node to the group
        /// </summary>
        /// <param name="_node">The node to add</param>
        public void AddNode(SimpleNode _node)
        {
            if (Title == "")
                Title = _node.Name + " et al.";

            _node.ParentNode = this;
            nodes.Add(_node);
            UpdateConnections();
        }

        /// <summary>
        /// Removes a node from the group
        /// </summary>
        /// <param name="_node">The node to remove</param>
        public void RemoveNode(SimpleNode _node)
        {
            _node.ParentNode = null;
            nodes.Remove(_node);
            UpdateConnections();
        }

        private void UpdateConnections()
        {
            connections.Clear();
            innerConnections.Clear();

            foreach (SimpleNode sn in nodes)
            {
                foreach (VizEdge edge in sn.Connections)
                {
                    SimpleNode otherNode = edge.Tail as SimpleNode;
                    if (otherNode == sn)
                        otherNode = edge.Head as SimpleNode;

                    if (nodes.Contains(otherNode))
                    {
                        if (edge.Head == sn)
                            innerConnections.Add(edge);
                    }
                    else
                    {
                        connections.Add(edge);
                    }
                }
            }
        }

        /// <summary>
        /// Returns the next valid point on the nodes boundingbox in direction of the edge
        /// </summary>
        /// <param name="_edge">Edge defining the direction</param>
        /// <returns>The nearest point on the boundingbox</returns>
        public override Point PointOnBox(VizEdge _edge)
        {
            if (this.NodesPosition.Count != 0)
            {
                Vector direction;

                BaseNode headNode = _edge.Head as BaseNode;
                if ((_edge.Head as SimpleNode).IsGrouped)
                    headNode = (_edge.Head as SimpleNode).ParentNode;

                BaseNode tailNode = _edge.Tail as BaseNode;
                if ((_edge.Tail as SimpleNode).IsGrouped)
                    tailNode = (_edge.Tail as SimpleNode).ParentNode;

                int index = 0;

                if (tailNode == this)
                {
                    direction = (tailNode.Pos - headNode.Pos);
                    index = Nodes.IndexOf(_edge.Tail as SimpleNode);
                }
                else
                {
                    direction = (headNode.Pos - tailNode.Pos);
                    index = Nodes.IndexOf(_edge.Head as SimpleNode);
                }

                if (index != -1 && (direction.X != 0.0 || direction.Y != 0.0))
                {
                    direction.Normalize();

                    //Hauptrichtung feststellen
                    if (Math.Abs(direction.X) > Math.Abs(direction.Y))
                        return this.Pos + new Vector(Math.Sign(-direction.X) * this.Size.Width / 2.0, this.NodesPosition[index].Y - this.Size.Height / 2.0);
                    else
                        return this.Pos + new Vector(this.NodesPosition[index].X - this.Size.Width / 2.0, Math.Sign(-direction.Y) * this.Size.Height / 2.0);
                }
                else
                    return this.Pos;
            }
            else
                return this.Pos;
        }
    }
}
