﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using GraphVizNet;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Xml;
using System.IO;
using Microsoft.Win32;

namespace Vis2NodeTrix
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        /// <summary>
        /// All nodes which are currently visible in the graph
        /// </summary>
        public ObservableCollection<BaseNode> Nodes { get; set; }
        private String inputDataFile = null;
        private Point boxSelectPoint;
        private Window boxSelectWindow;
        private Point panPoint;

        private const bool removeDoubleNodes = true;
        private const bool removeUnlinkedNodes = true;

        /// <summary>
        /// Initializes a new instance of the MainWindow class
        /// </summary>
        public MainWindow()
        {
            this.DataContext = this;

            InitializeComponent();

            Nodes = new ObservableCollection<BaseNode>();

            OnPropertyChanged("Nodes");

            this.Loaded += new RoutedEventHandler(MainWindow_Loaded);   
        }

        void MainWindow_Loaded(object sender, RoutedEventArgs e)
        {
            this.graphPresenter1.LayoutGraph(Nodes);
            this.graphPresenter1.UpdateVisuals();
            //OnPropertyChanged("Nodes");
        }

        /// <summary>
        /// Loads raw data into the visualizer and preprocesses them.
        /// </summary>
        /// <param name="doc">The xml document from which the data should be loaded</param>
        /// <param name="_removeDoubleNodes">Removes multiple nodes if true</param>
        /// <param name="_removeUnlinkedNodes">Removes nodes without links if true</param>
        /// <returns></returns>
        private List<SimpleNode> loadRawInputData(XmlDocument doc, bool _removeDoubleNodes = true, bool _removeUnlinkedNodes = true)
        {
            XmlNodeList xmlNodes = doc.GetElementsByTagName("node");

            List<SimpleNode> loadedNodes = new List<SimpleNode>();

            foreach (XmlNode n in xmlNodes)
            {
                SimpleNode newn = new SimpleNode();
                newn.Name = n.InnerText;

                if (!_removeDoubleNodes || !loadedNodes.Exists(x => x.Name == newn.Name))
                {
                    loadedNodes.Add(newn);
                }
            }

            XmlNodeList xmlLinks = doc.GetElementsByTagName("link");

            foreach (XmlNode n in xmlLinks)
            {
                SimpleNode s1 = loadedNodes.Find(x => x.Name == n.Attributes.Item(0).InnerText); //Node1
                SimpleNode s2 = loadedNodes.Find(x => x.Name == n.Attributes.Item(1).InnerText);
                if (s1 != null && s2 != null)
                    s1.AddOriginalNode(s2); //Node2
            }

            if(_removeUnlinkedNodes)
                loadedNodes.RemoveAll(x => x.Connections.Count == 0);

            return loadedNodes;
        }

        /// <summary>
        /// Saves the current graph in XML-format
        /// </summary>
        /// <param name="_fileName">Path to the destination File</param>
        private void saveCurrentState(String _fileName)
        {
            XmlDocument doc = new XmlDocument();
            
            XmlElement topNode = doc.CreateElement("NodeTrixSaveState");
            XmlAttribute inputData = doc.CreateAttribute("InputData");
            inputData.InnerText = this.inputDataFile;
            topNode.Attributes.Append(inputData);
            doc.AppendChild(topNode);

            XmlElement groupNodes = doc.CreateElement("groups");
            XmlElement singleNodes = doc.CreateElement("nodes");
            XmlElement edges = doc.CreateElement("edges");

            foreach (BaseNode n in Nodes)
            {
                if (n is GroupNode)
                {
                    XmlElement groupNode = doc.CreateElement("group");
                    XmlAttribute groupName = doc.CreateAttribute("name");
                    XmlAttribute groupTitle = doc.CreateAttribute("title");
                    XmlAttribute groupPosX = doc.CreateAttribute("posX");
                    XmlAttribute groupPosY = doc.CreateAttribute("posY");
                    GroupNode node = n as GroupNode;
                    groupName.InnerText = node.Name;
                    groupTitle.InnerText = node.Title;
                    groupPosX.InnerText = node.Pos.X.ToString();
                    groupPosY.InnerText = node.Pos.Y.ToString();
                    groupNode.Attributes.Append(groupName);
                    groupNode.Attributes.Append(groupTitle);
                    groupNode.Attributes.Append(groupPosX);
                    groupNode.Attributes.Append(groupPosY);
                    foreach (SimpleNode sn in node.Nodes)
                    {
                        XmlElement partNode = doc.CreateElement("packednode");
                        partNode.InnerText = sn.Name;
                        groupNode.AppendChild(partNode);

                        setEdgeTag(doc, edges, sn);
                    }
                    groupNodes.AppendChild(groupNode);
                }
                if (n is SimpleNode)
                {
                    XmlElement singleNode = doc.CreateElement("node");
                    XmlAttribute nodeName = doc.CreateAttribute("name");
                    XmlAttribute nodePosX = doc.CreateAttribute("posX");
                    XmlAttribute nodePosY = doc.CreateAttribute("posY");
                    SimpleNode node = n as SimpleNode;
                    nodeName.InnerText = node.Name;
                    nodePosX.InnerText = node.Pos.X.ToString();
                    nodePosY.InnerText = node.Pos.Y.ToString();
                    singleNode.Attributes.Append(nodeName);
                    singleNode.Attributes.Append(nodePosX);
                    singleNode.Attributes.Append(nodePosY);
                    singleNodes.AppendChild(singleNode);

                    setEdgeTag(doc, edges, node);
                }
            }

            XmlElement transforms = doc.CreateElement("transform");
            XmlElement translate = doc.CreateElement("translate");
            XmlAttribute translateX = doc.CreateAttribute("valX");
            translateX.InnerText = graphPresenter1.Translation.X.ToString();
            translate.Attributes.Append(translateX);
            XmlAttribute translateY = doc.CreateAttribute("valY");
            translateY.InnerText = graphPresenter1.Translation.Y.ToString();
            translate.Attributes.Append(translateY);
            XmlElement zoom = doc.CreateElement("zoom");
            XmlAttribute zoomVal = doc.CreateAttribute("value");
            zoomVal.InnerText = graphPresenter1.Zoom.ToString();
            zoom.Attributes.Append(zoomVal);
            transforms.AppendChild(translate);
            transforms.AppendChild(zoom);

            topNode.AppendChild(groupNodes);
            topNode.AppendChild(singleNodes);
            topNode.AppendChild(edges);
            topNode.AppendChild(transforms);

            doc.Save(_fileName);
        }

        private void setEdgeTag(XmlDocument doc,XmlElement edges,SimpleNode n)
        {
            foreach (VizEdge e in n.Connections)
            {
                if (e.Head.Name == n.Name)
                {
                    XmlElement edge = doc.CreateElement("edge");
                    XmlAttribute n1 = doc.CreateAttribute("node1");
                    XmlAttribute n2 = doc.CreateAttribute("node2");
                    n1.InnerText = n.Name;
                    n2.InnerText = e.Tail.Name;
                    String points = "";
                    foreach (Point p in e.Points)
                    {
                        if (points == "")
                        {
                            points += p.X.ToString() + ";" + p.Y.ToString();
                        }
                        else
                        {
                            points += ";" + p.X.ToString() + ";" + p.Y.ToString();
                        }
                    }
                    edge.InnerText = points;
                    edge.Attributes.Append(n1);
                    edge.Attributes.Append(n2);
                    edges.AppendChild(edge);
                }
            }
        }

        /// <summary>
        /// Loads a saved graph state.
        /// </summary>
        /// <param name="doc">The xml document from which the data should be loaded</param>
        private void loadState(XmlDocument doc)
        {
            String inputDataFile = doc.DocumentElement.Attributes.Item(0).InnerText;

            //XmlDocument docInput = new XmlDocument();
            //docInput.Load(inputDataFile);
            //List<SimpleNode> rawNodes = loadRawInputData(docInput, removeDoubleNodes, removeUnlinkedNodes);
            List<SimpleNode> nodes = new List<SimpleNode>();
            List<GroupNode> groupNodes = new List<GroupNode>();
            List<SimpleNode> singleNodes = new List<SimpleNode>();

            foreach (XmlNode n in doc.GetElementsByTagName("packednode"))
            {
                SimpleNode sn = new SimpleNode();
                sn.Name = n.InnerText;
                nodes.Add(sn);
            }
            foreach (XmlNode n in doc.GetElementsByTagName("node"))
            {
                SimpleNode sn = new SimpleNode();
                sn.Name = n.Attributes.Item(0).InnerText;
                nodes.Add(sn);
            }
            foreach (XmlNode e in doc.GetElementsByTagName("edge"))
            {
                VizEdge edge = new VizEdge();
                SimpleNode n1 = nodes.Find(x => x.Name == e.Attributes.Item(0).InnerText);
                SimpleNode n2 = nodes.Find(x => x.Name == e.Attributes.Item(1).InnerText);
                edge.Head = n1;
                edge.Tail = n2;
                String[] pointsArgs = e.InnerText.Split(';');
                edge.Points.Clear();
                for (int i = 0; i < pointsArgs.Length; i += 2)
                {
                    edge.Points.Add(new Point(Double.Parse(pointsArgs[i]), double.Parse(pointsArgs[i + 1])));
                }
                n1.Connections.Add(edge);
                n2.Connections.Add(edge);
            }

            foreach (XmlNode g in doc.GetElementsByTagName("group"))
            {
                GroupNode gn = new GroupNode();
                foreach (XmlNode n in g.ChildNodes)
                {
                    SimpleNode sn = nodes.Find(x => x.Name == n.InnerText);
                    gn.AddNode(sn);
                }
                gn.Name = g.Attributes.Item(0).InnerText;
                gn.Title = g.Attributes.Item(1).InnerText;
                gn.Pos = new Point(Double.Parse(g.Attributes.Item(2).InnerText), Double.Parse(g.Attributes.Item(3).InnerText));
                groupNodes.Add(gn);
            }
            foreach (XmlNode n in doc.GetElementsByTagName("node"))
            {
                SimpleNode sn = nodes.Find(x => x.Name == n.Attributes.Item(0).InnerText);
                sn.Pos = new Point(Double.Parse(n.Attributes.Item(1).InnerText), Double.Parse(n.Attributes.Item(2).InnerText));
                singleNodes.Add(sn);
            }

            Nodes.Clear();

            while (groupNodes.Count > 0 || singleNodes.Count > 0)
            {
                BaseNode node;
                
                if (singleNodes.Count > 0)
                {
                    node = singleNodes.First<SimpleNode>();
                    singleNodes.Remove((SimpleNode)node);
                }
                else
                {
                    node = groupNodes.First<GroupNode>();
                    groupNodes.Remove((GroupNode)node);
                }
                Nodes.Add(node);
            }

            XmlNode translate = doc.GetElementsByTagName("translate").Item(0);
            Point transPoint = new Point(Double.Parse(translate.Attributes.Item(0).InnerText), Double.Parse(translate.Attributes.Item(1).InnerText));
            graphPresenter1.Translation = transPoint;
            XmlNode zoom = doc.GetElementsByTagName("zoom").Item(0);
            double zoomVal = Double.Parse(zoom.Attributes.Item(0).InnerText);
            graphPresenter1.Zoom = zoomVal;
        }

        private void OnPropertyChanged(string _property)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(_property));
        }

        /// <summary>
        /// Event to be thrown if a property is changes
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        private void RelayoutButton_Click(object sender, RoutedEventArgs e)
        {
            relayout();
        }

        /// <summary>
        /// Auto-relayouts the graph
        /// </summary>
        private void relayout()
        {
            this.graphPresenter1.LayoutGraph(Nodes);
            this.graphPresenter1.UpdateVisuals();
        }

        private void SaveButton_Click(object sender, RoutedEventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.AddExtension = true;
            dlg.DefaultExt = ".xml";
            dlg.Filter = "XML File|*.xml|All Files|*.*";

            if (dlg.ShowDialog() == true)
                saveCurrentState(dlg.FileName);
        }

        private void LoadButton_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Title = "Load State";
            dlg.Filter = "XML File|*.xml|All Files|*.*";
            dlg.ValidateNames = true;

            if (dlg.ShowDialog() == true)
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(dlg.FileName);

                if (doc.DocumentElement.Name.Equals("NodeTrixSaveState"))
                {
                    loadState(doc);
                }
                else if (doc.DocumentElement.Name.Equals("NodeTrixInputData"))
                {
                    List<SimpleNode> snodes = loadRawInputData(doc, true, true);

                    Nodes.Clear();
                    graphPresenter1.Translation = new Point();
                    graphPresenter1.Zoom = 1.0;
                    foreach (SimpleNode n in snodes)
                        Nodes.Add(n);

                    OnPropertyChanged("Nodes");

                    relayout();
                    graphPresenter1.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
                    graphPresenter1.Arrange(new Rect(new Point(0,0), graphPresenter1.DesiredSize));
                    graphPresenter1.UpdateVisuals();
                }
            }
        }

        private void ScrollViewer_Drop(object sender, DragEventArgs e)
        {
            if (e.Data.GetData("SimpleNode") is SimpleNodePresenter)
            {
                SimpleNodePresenter np = e.Data.GetData("SimpleNode") as SimpleNodePresenter;
                Point p = e.GetPosition(graphPresenter1);
                p = new Point(p.X - (GraphPresenter.Border / 2) - np.dragPoint.X, p.Y - (GraphPresenter.Border / 2) - np.dragPoint.Y);
                np.Node.Pos = p;
            }
            if (e.Data.GetData("GroupNode") is GroupNodePresenter)
            {
                GroupNodePresenter np = e.Data.GetData("GroupNode") as GroupNodePresenter;
                Point p = e.GetPosition(graphPresenter1);
                p = new Point(p.X - (GraphPresenter.Border / 2) - np.dragPoint.X, p.Y - (GraphPresenter.Border / 2) - np.dragPoint.Y);
                np.Node.Pos = p;
            }
            if (e.Data.GetDataPresent("GroupNodeChild"))
            {
                SimpleNode n = e.Data.GetData("GroupNodeChild") as SimpleNode;
                GroupNode gn = n.ParentNode;
                Point p = e.GetPosition(graphPresenter1);
                p = new Point(p.X - (GraphPresenter.Border / 2), p.Y - (GraphPresenter.Border / 2));

                gn.RemoveNode(n);

                if (gn.Nodes.Count == 1)
                {
                    SimpleNode fn = gn.Nodes.First();
                    gn.RemoveNode(fn);
                    fn.Pos = gn.Pos;
                    graphPresenter1.Nodes.Add(fn);
                    graphPresenter1.Nodes.Remove(gn);
                }

                graphPresenter1.Nodes.Add(n);
                n.Pos = p;

                graphPresenter1.UpdateVisuals();
            }
        }

        private void ScrollViewer_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetData("SimpleNode") is SimpleNodePresenter)
            {
                SimpleNodePresenter np = e.Data.GetData("SimpleNode") as SimpleNodePresenter;

                Point p = e.GetPosition(graphPresenter1);
                p = new Point(p.X - (GraphPresenter.Border / 2) - np.dragPoint.X, p.Y - (GraphPresenter.Border / 2) - np.dragPoint.Y);

                np.Node.Pos = p;
                e.Handled = true;
            }
            if (e.Data.GetData("GroupNode") is GroupNodePresenter)
            {
                GroupNodePresenter np = e.Data.GetData("GroupNode") as GroupNodePresenter;

                Point p = e.GetPosition(graphPresenter1);
                p = new Point(p.X - (GraphPresenter.Border / 2) - np.dragPoint.X, p.Y - (GraphPresenter.Border / 2) - np.dragPoint.Y);

                np.Node.Pos = p;
            }
            if (e.Effects == DragDropEffects.None)
            {
                e.Handled = true;
            }
        }

        private void ScrollViewer_MouseWheel(object sender, MouseWheelEventArgs e)
        {
            Point mousePoint = e.GetPosition(graphPresenter1);
            Point oldPoint = graphPresenter1.RenderTransform.Transform(mousePoint);

            double ratio = 1 + (double)e.Delta / 2000.0;
            ratio *= ratio * graphPresenter1.Zoom;
            graphPresenter1.Zoom = ratio;

            Point newPoint = graphPresenter1.RenderTransform.Transform(mousePoint);

            Point translationPoint = new Point(oldPoint.X - newPoint.X, oldPoint.Y - newPoint.Y);

            graphPresenter1.Translation = new Point(graphPresenter1.Translation.X + translationPoint.X, graphPresenter1.Translation.Y + translationPoint.Y);

            Console.WriteLine("Difference: {0}  {1}", translationPoint.X, translationPoint.Y);
            e.Handled = true;
        }

        private void ScrollViewer_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (Keyboard.IsKeyDown(Key.LeftCtrl))
            {
                boxSelectPoint = e.GetPosition(graphPresenter1);
                showBoxSelectWindow();
                panPoint = new Point();
            }
            else
            {
                panPoint = e.GetPosition(this);
                boxSelectPoint = new Point();
            }
        }

        /// <summary>
        /// Creates and activates the boxselection overlay window.
        /// </summary>
        private void showBoxSelectWindow()
        {
            boxSelectWindow = new Window();
            boxSelectWindow.WindowStyle = WindowStyle.None;
            boxSelectWindow.AllowsTransparency = true;
            boxSelectWindow.AllowDrop = false;
            boxSelectWindow.Background = Brushes.Transparent;
            boxSelectWindow.IsHitTestVisible = true;
            boxSelectWindow.SizeToContent = SizeToContent.WidthAndHeight;
            boxSelectWindow.Topmost = true;
            boxSelectWindow.ShowInTaskbar = false;
            boxSelectWindow.PreviewMouseMove += new MouseEventHandler(ScrollViewer_PreviewMouseMove);

            Rectangle r = new Rectangle();
            DoubleCollection dc = new DoubleCollection();
            dc.Add(2.0);dc.Add(1.0);
            r.StrokeDashArray = dc;
            r.Fill = new SolidColorBrush(Colors.LightBlue);
            r.Opacity = 0.5;
            boxSelectWindow.Content = r;

            Point pos = graphPresenter1.PointToScreen(boxSelectPoint);
            boxSelectWindow.Left = pos.X;
            boxSelectWindow.Top = pos.Y;

            boxSelectWindow.Show();
        }

        private void ScrollViewer_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (e.LeftButton == MouseButtonState.Pressed && panPoint.X != 0.0 && panPoint.Y != 0.0 && !(boxSelectWindow != null && boxSelectWindow.IsVisible))
            {
                Point akPos = e.GetPosition(this);
                double difX = akPos.X - panPoint.X + graphPresenter1.Translation.X;
                double difY = akPos.Y - panPoint.Y + graphPresenter1.Translation.Y;

                graphPresenter1.Translation = new Point(difX, difY);
                panPoint = akPos;
            }
            else if (boxSelectPoint.X != 0.0 && boxSelectPoint.Y != 0.0)
            {
                if (e.LeftButton == MouseButtonState.Released)
                {
                    boxSelectPoint = new Point();
                    boxSelectWindow.Close();
                }

                Point akPos = e.GetPosition(graphPresenter1);
                Point akPosScreen = graphPresenter1.PointToScreen(akPos);
                Point boxSelectPointScreen = graphPresenter1.PointToScreen(boxSelectPoint);

                boxSelectWindow.Left = Math.Min(akPosScreen.X, boxSelectPointScreen.X);
                boxSelectWindow.Top = Math.Min(akPosScreen.Y, boxSelectPointScreen.Y);
                (boxSelectWindow.Content as Rectangle).Width = Math.Abs(akPosScreen.X - boxSelectPointScreen.X);
                (boxSelectWindow.Content as Rectangle).Height = Math.Abs(akPosScreen.Y - boxSelectPointScreen.Y);
            }
        }

        private void ScrollViewer_PreviewMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            panPoint = new Point();

            if (boxSelectPoint != new Point())
            {
                boxSelectWindow.Close();
                Point endPoint = e.GetPosition(graphPresenter1);
                Point centerPoint = new Point();
                List<BaseNode> selectedNodes = new List<BaseNode>();
                double minX = Math.Min(boxSelectPoint.X, endPoint.X) - GraphPresenter.Border / 2.0;
                double maxX = Math.Max(boxSelectPoint.X, endPoint.X) - GraphPresenter.Border / 2.0;
                double minY = Math.Min(boxSelectPoint.Y, endPoint.Y) - GraphPresenter.Border / 2.0;
                double maxY = Math.Max(boxSelectPoint.Y, endPoint.Y) - GraphPresenter.Border / 2.0;

                foreach (BaseNode n in graphPresenter1.Nodes)
                {
                    if (n.Pos.X - n.Size.Width / 2 >= minX &&
                        n.Pos.X + n.Size.Width / 2 <= maxX &&
                        n.Pos.Y - n.Size.Height / 2 >= minY &&
                        n.Pos.Y + n.Size.Height / 2 <= maxY)
                    {
                        selectedNodes.Add(n);
                        centerPoint += new Vector(n.Pos.X, n.Pos.Y);
                    }
                }

                if (selectedNodes.Count < 2)
                {
                    boxSelectPoint = new Point();
                    return;
                }

                centerPoint = new Point(centerPoint.X / (double)selectedNodes.Count, centerPoint.Y / (double)selectedNodes.Count);

                GroupNode groupnode = new GroupNode();
                groupnode.Pos = centerPoint;
                double[] values = {centerPoint.X,centerPoint.Y};
                foreach (BaseNode n in selectedNodes)
                {
                    n.OnAnimation(AnimationType.TRANSLATE_TO_MATRIX, values,groupnode);
                }

                boxSelectPoint = new Point();
            }
        }

        private RenderTargetBitmap RenderVisual(Visual _visual)
        {
            if (MainWindow.renderTarget == null)
            {
                MainWindow.renderTarget = new RenderTargetBitmap(
                    (int)graphPresenter1.ActualWidth,
                    (int)graphPresenter1.ActualHeight,
                    96,
                    96,
                    PixelFormats.Default);
            }
            renderTarget.Render(_visual);

            return renderTarget;
        }

        private static RenderTargetBitmap renderTarget;

        private void SaveImage_Click(object sender, RoutedEventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.AddExtension = true;
            dlg.DefaultExt = ".png";
            dlg.Filter = "PNG File|*.png|All Files|*.*";

            if (dlg.ShowDialog() == true)
            {
                Transform t = graphPresenter1.RenderTransform;
                graphPresenter1.RenderTransform = new TransformGroup();
                graphPresenter1.UpdateLayout();
//                graphPresenter1.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
//                graphPresenter1.Arrange(new Rect(0, 0, this.DesiredSize.Width, this.DesiredSize.Height));

                RenderTargetBitmap rtb = RenderVisual((Visual)graphPresenter1);                
                PngBitmapEncoder encoder = new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(rtb));

                Stream fs = new MemoryStream();//File.Open(dlg.FileName + ".tmp", FileMode.OpenOrCreate);

                encoder.Save(fs);
                encoder = null;

                fs.Flush();
                fs.Position = 0;

                System.Drawing.Image bm = System.Drawing.Image.FromStream(fs);
                fs.Close();
                fs.Dispose();
                fs = null;

                int minX = (int)Math.Max(0.0, graphPresenter1.minCoords.X - 1);
                int minY = (int)Math.Max(0.0, graphPresenter1.minCoords.Y - 1);
                System.Drawing.Bitmap newbm = new System.Drawing.Bitmap((int)graphPresenter1.ActualWidth - minX + 50, (int)graphPresenter1.ActualHeight - minY + 50);

                using (System.Drawing.Graphics g = System.Drawing.Graphics.FromImage(newbm))
                {
                    g.DrawImage(bm, 50, 50, new System.Drawing.Rectangle(minX, minY, bm.Width - minX, bm.Height - minY), System.Drawing.GraphicsUnit.Pixel);
                    g.Flush();
                    g.Dispose();
                }

                newbm.Save(dlg.FileName);
                bm.Dispose();
                newbm.Dispose();
                bm = null;
                newbm = null;

                graphPresenter1.RenderTransform = t;
                graphPresenter1.UpdateLayout();

                rtb.Clear();
                rtb = null;

                GC.Collect();
                GC.WaitForPendingFinalizers();
                GC.Collect();

                //graphPresenter1.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
                //graphPresenter1.Arrange(new Rect(0, 0, this.DesiredSize.Width, this.DesiredSize.Height));
            }
        }

        private void ExitButton_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void AboutButton_Click(object sender, RoutedEventArgs e)
        {
            MessageBox.Show("Vis2NodeTrix\n\nAuthors:\nKlemens Jahrmann\nBernhard Steiner\n");
        }
    }
}
